import bpy

from .. import utils


class POWERPROPS_OT_copy_property(bpy.types.Operator):
    bl_idname = 'powerprops.copy_property'
    bl_label = 'Copy Property'
    bl_description = '.\n'.join((
        'Copy property from the active object to all selected objects',
        'Shift click to only copy to objects that already have this property',
        'Ctrl click to recreate drivers that depend on this property',
        'Alt click to only copy property settings',
    ))
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    data_path: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    property_name: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    batch: bpy.props.BoolProperty(options={'HIDDEN', 'SKIP_SAVE'})

    only_existing: bpy.props.BoolProperty(
        name='Only Existing',
        description='Only copy to objects that already have this property',
        default=False,
    )

    dependent_drivers: bpy.props.BoolProperty(
        name='Dependent Drivers',
        description='Recreate drivers that depend on this property',
        default=False,
    )

    only_settings: bpy.props.BoolProperty(
        name='Only Settings',
        description='Only copy property settings, leave values as they are or use defaults',
        default=False,
    )

    def invoke(self, context: bpy.types.Context, event: bpy.types.Event) -> set[str]:
        self.only_existing = event.shift
        self.dependent_drivers = event.ctrl
        self.only_settings = event.alt

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context: bpy.types.Context, event: bpy.types.Event):
        return self.execute(context)

    def execute(self, context: bpy.types.Context) -> set:
        source_data: bpy.types.ID = context.path_resolve(self.data_path)

        if self.batch:
            wm_props = context.window_manager.powerprops
            property_names = set(p.name for p in wm_props.properties if p.select)
            property_names = property_names.intersection(source_data.keys())
        else:
            property_names = [self.property_name]

        data_blocks = set()
        drivers = 0

        for property_name in property_names:
            skip = {source_data, None}

            for object in context.selected_objects:
                with context.temp_override(active_object=object):
                    target_data: bpy.types.ID = context.path_resolve(self.data_path)

                if target_data in skip:
                    continue

                skip.add(target_data)

                if self.only_existing:
                    if property_name not in target_data:
                        continue

                if self.only_settings:
                    if not utils.property.copy_type(source_data, target_data, property_name):
                        continue
                else:
                    target_data[property_name] = source_data[property_name]

                utils.property.copy_ui(source_data, target_data, property_name)
                if self.dependent_drivers:
                    drivers += utils.property.copy_drivers(source_data, target_data, property_name)

                data_blocks.add(target_data)

        utils.common.tag_redraw(context)
        for data_block in data_blocks:
            utils.common.update_tag(context, data_block)

        info_a = f'{len(property_names)} properties' if self.batch else f'"{self.property_name}"'
        info_b = f'Copied {info_a} to {len(data_blocks)} data blocks'
        self.report({'INFO'}, f'{info_b} and copied {drivers} drivers' if self.dependent_drivers else info_b)
        return {'FINISHED'}
