import bpy

from .. import utils

WM_OT_properties_edit_mixin = type('', (), dict(bpy.types.WM_OT_properties_edit.__dict__))


class POWERPROPS_OT_edit_property(bpy.types.Operator, WM_OT_properties_edit_mixin):
    bl_idname = 'powerprops.edit_property'
    bl_label = 'Edit Property'
    bl_description = '.\n'.join((
        'Edit details about the property',
        'Shift click to copy your changes to all selected objects',
    ))
    bl_options = {'UNDO', 'INTERNAL'}

    data_path: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    property_name: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})

    edit_selected: bpy.props.BoolProperty(
        name='Edit Selected',
        description='Edit property on all selected objects',
        default=False,
    )

    def draw(self, context: bpy.types.Context):
        super().draw(context)
        self.layout.prop(self, 'edit_selected')

    def invoke(self, context: bpy.types.Context, event: bpy.types.Event) -> set[str]:
        self.edit_selected = event.shift
        return super().invoke(context, event)

    def execute(self, context: bpy.types.Context) -> set[str]:
        source_data: bpy.types.ID = context.path_resolve(self.data_path)

        name_old: str = self._old_prop_name[0]
        name_new: str = self.property_name

        result = super().execute(context)
        if 'FINISHED' not in result:
            return result

        objects = {context.active_object}
        data_blocks = {source_data}

        if self.edit_selected:
            for object in context.selected_objects:
                with context.temp_override(active_object=object):
                    target_data: bpy.types.ID = context.path_resolve(self.data_path)

                if not target_data or target_data in data_blocks or name_old not in target_data:
                    continue

                if name_old != name_new:
                    utils.property.update_property(target_data, name_old, name_new)

                    objects.add(object)
                    data_blocks.add(target_data)

                if utils.property.copy_type(source_data, target_data, name_new):
                    utils.property.copy_ui(source_data, target_data, name_new)

                    objects.add(object)
                    data_blocks.add(target_data)

        if name_old != name_new:
            utils.property.update_selection(context, name_old, name_new)

            for object in objects:
                utils.property.update_particles(object, name_old, name_new)
                utils.property.update_materials(object, name_old, name_new, 'OBJECT')

            for data_block in data_blocks:
                utils.property.update_all_drivers(data_block, name_old, name_new)

        utils.common.tag_redraw(context)
        for data_block in data_blocks:
            utils.common.update_tag(context, data_block)

        info = f'{len(data_blocks)} data blocks' if self.edit_selected else f'"{source_data.name}"'
        self.report({'INFO'}, f'Edited "{name_new}" on {info}')
        return {'FINISHED'}
