import random

import bpy

from .. import utils


class POWERPROPS_OT_randomize_property(bpy.types.Operator):
    bl_idname = 'powerprops.randomize_property'
    bl_label = 'Randomize Property'
    bl_description = '.\n'.join((
        'Randomize property value',
        'Shift click to randomize the property on all selected objects',
    ))
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    data_path: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    property_name: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})

    on_selected: bpy.props.BoolProperty(
        name='On Selected',
        description='Randomize property on all selected objects',
        default=False,
    )

    seed: bpy.props.IntProperty(name='Seed', description='Seed for the random number generator')
    mask: bpy.props.BoolVectorProperty(name='Mask', size=32)

    int_min: bpy.props.IntVectorProperty(name='Min', size=32)
    int_max: bpy.props.IntVectorProperty(name='Max', size=32)

    float_min: bpy.props.FloatVectorProperty(name='Min', size=32)
    float_max: bpy.props.FloatVectorProperty(name='Max', size=32)

    def draw(self, context: bpy.types.Context):
        layout = self.layout

        layout.prop(self, 'on_selected')
        layout.separator()
        layout.prop(self, 'seed')
        layout.separator()

        row = layout.row()
        if isinstance(self.value, list):
            rowsub = row.row(align=True)

            col = rowsub.column(align=True)
            for index in range(len(self.value)):
                col.prop(self, 'mask', text='', index=index)

            col = rowsub.column(align=True)
            for index in range(len(self.value)):
                colsub = col.row(align=True)
                colsub.enabled = self.mask[index]
                colsub.prop(self, self.min_name, text='Min' if index == 0 else ' ', index=index)

            col = row.row().column(align=True)
            for index in range(len(self.value)):
                colsub = col.row(align=True)
                colsub.enabled = self.mask[index]
                colsub.prop(self, self.max_name, text='Max' if index == 0 else ' ', index=index)

        else:
            row.prop(self, self.min_name, text='Min', index=0)
            row.prop(self, self.max_name, text='Max', index=0)

    def invoke(self, context: bpy.types.Context, event: bpy.types.Event) -> set[str]:
        active_data: bpy.types.ID = context.path_resolve(self.data_path)

        self.on_selected = event.shift
        self.seed = random.randint(-10000, 10000)
        self.mask = [True] * 32

        self.value = utils.property.to_builtin(active_data[self.property_name])
        self.type = type(self.value[0] if isinstance(self.value, list) else self.value)
        self.ui: dict = active_data.id_properties_ui(self.property_name).as_dict()

        self.min_name = 'int_min' if self.type == int else 'float_min'
        self.max_name = 'int_max' if self.type == int else 'float_max'

        self.min = getattr(self, self.min_name)
        self.max = getattr(self, self.max_name)

        for index in range(32):
            self.min[index] = self.ui['soft_min']
            self.max[index] = self.ui['soft_max']

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context: bpy.types.Context, event: bpy.types.Event):
        return self.execute(context)

    def execute(self, context: bpy.types.Context) -> set:
        active_data: bpy.types.ID = context.path_resolve(self.data_path)

        objects = {context.active_object}
        if self.on_selected:
            objects.update(context.selected_objects)

        seed = self.seed
        data_blocks = set()

        for object in objects:
            with context.temp_override(active_object=object):
                object_data: bpy.types.ID = context.path_resolve(self.data_path)

            if not object_data or self.property_name not in object_data:
                continue
            if not utils.property.compare_type(self.value, object_data[self.property_name], length=True):
                continue

            if isinstance(self.value, list):
                for index in range(len(self.value)):
                    if self.mask[index]:
                        object_data[self.property_name][index] = utils.common.seeded_random(seed, self.min[index], self.max[index])
                    seed += 1

            else:
                object_data[self.property_name] = utils.common.seeded_random(seed, self.min[0], self.max[0])
                seed += 1

            data_blocks.add(object_data)

        utils.common.tag_redraw(context)
        for data_block in data_blocks:
            utils.common.update_tag(context, data_block)

        info = f'{len(data_blocks)} data blocks' if self.on_selected else f'"{active_data.name}"'
        self.report({'INFO'}, f'Randomized "{self.property_name}" on {info}')
        return {'FINISHED'}
