import bpy

from .. import utils


class POWERPROPS_OT_remove_property(bpy.types.Operator):
    bl_idname = 'powerprops.remove_property'
    bl_label = 'Remove Property'
    bl_description = '.\n'.join((
        'Remove property from the active object',
        'Shift click to remove it from all selected objects',
        'Ctrl click to remove drivers that depend on this property',
    ))
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    data_path: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    property_name: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    batch: bpy.props.BoolProperty(options={'HIDDEN', 'SKIP_SAVE'})

    from_selected: bpy.props.BoolProperty(
        name='From Selected',
        description='Remove property from all selected objects',
        default=False,
    )

    dependent_drivers: bpy.props.BoolProperty(
        name='Dependent Drivers',
        description='Remove drivers that depend on this property',
        default=False,
    )

    def invoke(self, context: bpy.types.Context, event: bpy.types.Event) -> set[str]:
        self.from_selected = event.shift
        self.dependent_drivers = event.ctrl

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context: bpy.types.Context, event: bpy.types.Event):
        return self.execute(context)

    def execute(self, context: bpy.types.Context) -> set:
        active_data: bpy.types.ID = context.path_resolve(self.data_path)

        if self.batch:
            wm_props = context.window_manager.powerprops
            property_names = set(p.name for p in wm_props.properties if p.select)
            property_names = property_names.intersection(active_data.keys())
        else:
            property_names = [self.property_name]

        objects = {context.active_object}
        if self.from_selected:
            objects.update(context.selected_objects)

        data_blocks = set()
        drivers = 0

        for property_name in property_names:
            for object in objects:
                with context.temp_override(active_object=object):
                    object_data: bpy.types.ID = context.path_resolve(self.data_path)

                if object_data and property_name in object_data:
                    if self.dependent_drivers:
                        for fcurve, _ in utils.property.find_all_drivers(object_data, property_name):
                            fcurve.id_data.driver_remove(fcurve.data_path, fcurve.array_index)
                            drivers += 1

                    del object_data[property_name]
                    data_blocks.add(object_data)

        utils.common.tag_redraw(context)
        for data_block in data_blocks:
            utils.common.update_tag(context, data_block)

        info_a = f'{len(property_names)} properties' if self.batch else f'"{self.property_name}"'
        info_b = f'{len(data_blocks)} data blocks' if self.from_selected else f'"{active_data.name}"'
        info_c = f'Removed {info_a} from {info_b}'
        self.report({'INFO'}, f'{info_c} and removed {drivers} drivers' if self.dependent_drivers else info_c)
        return {'FINISHED'}
