import bpy

from .. import utils


class POWERPROPS_OT_select_by_property(bpy.types.Operator):
    bl_idname = 'powerprops.select_by_property'
    bl_label = 'Select by Property'
    bl_description = 'Select by property. Shift click to compare value'
    bl_description = '.\n'.join((
        'Select objects with this property',
        'Shift click to take values into account',
    ))
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    data_path: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    property_name: bpy.props.StringProperty(options={'HIDDEN', 'SKIP_SAVE'})
    batch: bpy.props.BoolProperty(options={'HIDDEN', 'SKIP_SAVE'})

    compare_value: bpy.props.BoolProperty(
        name='Compare Value',
        description='Only select objects if the property value matches',
        default=False,
    )

    def invoke(self, context: bpy.types.Context, event: bpy.types.Event) -> set[str]:
        self.compare_value = event.shift

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context: bpy.types.Context, event: bpy.types.Event):
        return self.execute(context)

    def execute(self, context: bpy.types.Context) -> set:
        source_data: bpy.types.ID = context.path_resolve(self.data_path)

        if self.batch:
            wm_props = context.window_manager.powerprops
            property_names = set(p.name for p in wm_props.properties if p.select)
            property_names = property_names.intersection(source_data.keys())
        else:
            property_names = [self.property_name]

        objects = set(context.selectable_objects)

        for property_name in property_names:
            source_value = utils.property.to_builtin(source_data[property_name])

            for object in objects.copy():
                with context.temp_override(active_object=object):
                    target_data: bpy.types.ID = context.path_resolve(self.data_path)

                if not target_data or property_name not in target_data:
                    objects.discard(object)

                elif self.compare_value:
                    target_value = utils.property.to_builtin(target_data[property_name])

                    if source_value != target_value:
                        objects.discard(object)

        for object in context.selectable_objects:
            object.select_set(object in objects)

        info = f'{len(property_names)} properties' if self.batch else f'"{self.property_name}"'
        self.report({'INFO'}, f'Selected {len(objects)} objects with {info}')
        return {'FINISHED'}
