import ast
import typing

import bpy

from .. import utils


def copy_type(source_data: bpy.types.ID, target_data: bpy.types.ID, property_name: str) -> bool:
    try:
        if property_name not in target_data:
            source_ui = source_data.id_properties_ui(property_name)
            target_data[property_name] = source_ui.as_dict()['default']

        source_value = to_builtin(source_data[property_name])
        target_value = to_builtin(target_data[property_name])
        target_data[property_name] = cast_type(source_value, target_value)
    except:
        return False
    else:
        return True


def copy_ui(source_data: bpy.types.ID, target_data: bpy.types.ID, property_name: str) -> bool:
    try:
        source_ui = source_data.id_properties_ui(property_name)
        target_ui = target_data.id_properties_ui(property_name)
        target_ui.update_from(source_ui)
    except:
        return False
    else:
        return True


def copy_drivers(source_data: bpy.types.ID, target_data: bpy.types.ID, property_name: str):
    drivers = 0

    for source_fcurve, _ in find_drivers(source_data, source_data, property_name):
        try:
            if hasattr(source_data[property_name], '__len__'):
                target_fcurve = target_data.driver_add(source_fcurve.data_path, source_fcurve.array_index)
            else:
                target_fcurve = target_data.driver_add(source_fcurve.data_path)

            target_fcurve.driver.type = source_fcurve.driver.type
            target_fcurve.driver.use_self = source_fcurve.driver.use_self
            target_fcurve.driver.expression = source_fcurve.driver.expression

            for source_variable in source_fcurve.driver.variables:
                target_variable = target_fcurve.driver.variables.new()
                target_variable.name = source_variable.name
                target_variable.type = source_variable.type

                for source_target, target_target in zip(source_variable.targets, target_variable.targets):
                    target_target.id_type = source_target.id_type
                    target_target.data_path = source_target.data_path

                    if source_target.id is source_data:
                        target_target.id = target_data
                    else:
                        target_target.id = source_target.id

                    target_target.bone_target = source_target.bone_target
                    target_target.rotation_mode = source_target.rotation_mode
                    target_target.transform_type = source_target.transform_type
                    target_target.transform_space = source_target.transform_space

        except:
            try:
                target_data.driver_remove(target_fcurve.data_path, target_fcurve.array_index)
            except:
                pass
        else:
            drivers += 1

    return drivers


def to_builtin(value: typing.Any) -> typing.Any | dict | list:
    if hasattr(value, 'to_dict'):
        value = value.to_dict()
    elif hasattr(value, 'to_list'):
        value = value.to_list()

    return value


def compare_type(a: typing.Any, b: typing.Any, length: bool = False) -> bool:
    a = to_builtin(a)
    b = to_builtin(b)

    if type(a) != type(b):
        return False

    if isinstance(a, list):
        if type(a[0]) != type(b[0]):
            return False

        if length and len(a) != len(b):
            return False

    return True


def cast_type(source: typing.Any, target: typing.Any) -> typing.Any:
    if isinstance(source, str):
        return type(source)(target)

    if isinstance(target, str):
        target = ast.literal_eval(target)

    if isinstance(source, list):
        if not isinstance(target, list):
            target = [target] * len(source)

        for index in range(len(target), len(source)):
            target.append(type(source[index])())

        return [type(s)(t) for s, t in zip(source, target)]

    if isinstance(target, list):
        target = target[0] if target else type(source)()

    return type(source)(target)


def update_property(data: bpy.types.ID, name_old: str, name_new: str):
    data[name_new] = data[name_old]

    try:
        ui_old = data.id_properties_ui(name_old)
        ui_new = data.id_properties_ui(name_new)
        ui_new.update_from(ui_old)
    except:
        pass

    del data[name_old]


def update_selection(context: bpy.types.Context, name_old: str, name_new: str):
    wm_props = context.window_manager.powerprops
    property_old = wm_props.ensure_property(name_old)
    property_new = wm_props.ensure_property(name_new)
    property_new.select = property_old.select


def update_particles(object: bpy.types.Object, name_old: str, name_new: str):
    for particle_system in object.particle_systems:
        if particle_system.settings.render_type == 'OBJECT':
            if particle_system.settings.instance_object:
                update_materials(particle_system.settings.instance_object, name_old, name_new, 'INSTANCER')

        elif particle_system.settings.render_type == 'COLLECTION':
            if particle_system.settings.instance_collection:
                for object in particle_system.settings.instance_collection.all_objects:
                    update_materials(object, name_old, name_new, 'INSTANCER')


def update_materials(object: bpy.types.Object, name_old: str, name_new: str, attribute_type: str):
    for material_slot in object.material_slots:
        if material_slot.material and material_slot.material.node_tree:
            update_shader(material_slot.material.node_tree, name_old, name_new, attribute_type)


def update_shader(node_tree: bpy.types.ShaderNodeTree, name_old: str, name_new: str, attribute_type: str):
    for node in node_tree.nodes:
        if isinstance(node, bpy.types.ShaderNodeAttribute):
            if node.attribute_type == attribute_type:
                if node.attribute_name == name_old:
                    node.attribute_name = name_new

        elif isinstance(node, bpy.types.ShaderNodeGroup):
            if node.node_tree:
                update_shader(node.node_tree, name_old, name_new, attribute_type)


def update_all_drivers(target_data: bpy.types.ID, name_old: str, name_new: str):
    path_old = f'["{bpy.utils.escape_identifier(name_old)}"]'
    path_new = f'["{bpy.utils.escape_identifier(name_new)}"]'

    for _, targets in find_all_drivers(target_data, name_old):
        for target in targets:
            target.data_path = target.data_path.replace(path_old, path_new)


def find_all_drivers(target_data: bpy.types.ID, property_name: str) -> list[tuple[bpy.types.FCurve, list[bpy.types.DriverTarget]]]:
    drivers = []

    for source_data in utils.common.get_data_blocks():
        drivers.extend(find_drivers(target_data, source_data, property_name))

    return drivers


def find_drivers(target_data: bpy.types.ID, source_data: bpy.types.ID, property_name: str) -> list[tuple[bpy.types.FCurve, list[bpy.types.DriverTarget]]]:
    drivers = []

    animation = getattr(source_data, 'animation_data', None)
    property_path = f'["{bpy.utils.escape_identifier(property_name)}"]'

    if isinstance(animation, bpy.types.AnimData):
        for fcurve in animation.drivers:
            targets = []

            for variable in fcurve.driver.variables:
                for target in variable.targets:
                    if target.id is target_data and property_path in target.data_path:
                        targets.append(target)

            if targets:
                drivers.append((fcurve, targets))

    return drivers


def to_ui_dict(property: bpy.types.Property) -> dict:
    ui_data = {'description': property.description}

    if property.type == 'BOOLEAN':
        ui_data['min'] = 0
        ui_data['max'] = 1

    elif property.type in ('INT', 'FLOAT'):
        ui_data['min'] = property.hard_min
        ui_data['max'] = property.hard_max
        ui_data['soft_min'] = property.soft_min
        ui_data['soft_max'] = property.soft_max
        ui_data['step'] = property.step

    if property.type == 'FLOAT':
        ui_data['precision'] = property.precision

    if property.is_array:
        ui_data['default'] = property.default_array
    else:
        ui_data['default'] = property.default

    if property.subtype in ('EULER', 'QUATERNION', 'COLOR', 'COLOR_GAMMA'):
        ui_data['subtype'] = property.subtype
    elif property.subtype == 'ANGLE':
        ui_data['subtype'] = 'EULER'

    return ui_data


def drive(source_data: bpy.types.ID, target_data: bpy.types.ID, property: bpy.types.Property, index: int = -1):
    source_path = utils.property.get_path(source_data, property)
    target_path = f'["{bpy.utils.escape_identifier(property.name)}"]'

    fcurve = source_data.driver_add(source_path, index)
    fcurve.driver.expression = 'var'

    variable = fcurve.driver.variables.new()
    variable.name = 'var'
    variable.type = 'SINGLE_PROP'

    target = variable.targets[0]
    target.id_type = utils.common.get_data_type(target_data)
    target.id = target_data

    if index == -1:
        target.data_path = target_path
    else:
        target.data_path = f'{target_path}[{index}]'


def get_path(data: bpy.types.ID, property: bpy.types.Property) -> str:
    if is_rna(data, property):
        return property.identifier
    else:
        return f'["{bpy.utils.escape_identifier(property.identifier)}"]'


def is_rna(data: bpy.types.ID, property: bpy.types.Property) -> bool:
    return property in data.bl_rna.properties.values()
